@@ -1,6 +1,7 @@
 {-# LANGUAGE TemplateHaskell #-}
 {-# LANGUAGE CPP #-}
 {-# LANGUAGE OverloadedStrings #-}
+{-# LANGUAGE RankNTypes #-}
 {-# OPTIONS_GHC -fno-warn-orphans -fno-warn-missing-fields #-}
 -- | This module provides utilities for creating backends. Regular users do not
 -- need to use this module.
@@ -15,8 +16,12 @@ module Database.Persist.TH
     , persistFile
       -- * Turn @EntityDef@s into types
     , mkPersist
-    , MkPersistSettings (..)
+    , MkPersistSettings
+    , mpsBackend
+    , mpsGeneric
+    , mkPersistSettings
     , sqlSettings
+    , sqlOnlySettings
       -- * Various other TH functions
     , mkMigrate
     , mkSave
@@ -34,8 +39,8 @@ import Database.Persist.Quasi
 import Database.Persist.Store
 import Database.Persist.Query.Internal
 import Database.Persist.GenericSql (Migration, SqlPersist, migrate)
-import Database.Persist.Util (nullable)
-import Database.Persist.TH.Library (apE)
+import Database.Persist.GenericSql.Raw (SqlBackend)
+import Database.Persist.Util (nullable, IsNullable(..), WhyNullable(..))
 import Language.Haskell.TH.Quote
 import Language.Haskell.TH.Syntax
 import Data.Char (toLower, toUpper)
@@ -94,9 +99,9 @@ persistFile = persistFileWith upperCaseSettings
 -- 'EntityDef's. Works well with the persist quasi-quoter.
 mkPersist :: MkPersistSettings -> [EntityDef] -> Q [Dec]
 mkPersist mps ents = do
-    x <- fmap mconcat $ mapM persistFieldFromEntity ents
+    x <- fmap mconcat $ mapM (persistFieldFromEntity mps) ents
     y <- fmap mconcat $ mapM (mkEntity mps) ents
-    z <- fmap mconcat $ mapM mkJSON ents
+    z <- fmap mconcat $ mapM (mkJSON mps) ents
     return $ mconcat [x, y, z]
 
 -- | Settings to be passed to the 'mkPersist' function.
@@ -107,13 +112,29 @@ data MkPersistSettings = MkPersistSettings
     -- When generating data types, each type is given a generic version- which
     -- works with any backend- and a type synonym for the commonly used
     -- backend. This is where you specify that commonly used backend.
+    , mpsGeneric :: Bool
+    -- ^ Create generic types that can be used with multiple backends. Good for
+    -- reusable code, but makes error messages harder to understand. Default:
+    -- True.
+    }
+
+-- | Create an @MkPersistSettings@ with default values.
+mkPersistSettings :: Type -- ^ Value for 'mpsBackend'
+                  -> MkPersistSettings
+mkPersistSettings t = MkPersistSettings
+    { mpsBackend = t
+    , mpsGeneric = True -- FIXME switch default to False in the future
     }
 
 -- | Use the 'SqlPersist' backend.
 sqlSettings :: MkPersistSettings
-sqlSettings = MkPersistSettings
-    { mpsBackend = ConT ''SqlPersist
-    }
+sqlSettings = mkPersistSettings $ ConT ''SqlBackend
+
+-- | Same as 'sqlSettings', but set 'mpsGeneric' to @False@.
+--
+-- Since 1.1.1
+sqlOnlySettings :: MkPersistSettings
+sqlOnlySettings = sqlSettings { mpsGeneric = False }
 
 recName :: Text -> Text -> Text
 recName dt f = lowerFirst dt ++ upperFirst f
@@ -130,79 +151,123 @@ upperFirst t =
         Just (a, b) -> cons (toUpper a) b
         Nothing -> t
 
-dataTypeDec :: EntityDef -> Dec
-dataTypeDec t =
-    DataD [] nameG [KindedTV backend monadTransKind] [RecC name cols]
+dataTypeDec :: MkPersistSettings -> EntityDef -> Dec
+dataTypeDec mps t =
+    DataD [] nameFinal paramsFinal constrs
     $ map (mkName . unpack) $ entityDerives t
   where
-    monadKind = StarK `ArrowK` StarK
-    monadTransKind = monadKind `ArrowK` monadKind
     mkCol x (FieldDef n _ ty as) =
         (mkName $ unpack $ recName x $ unHaskellName n,
          NotStrict,
-         pairToType backend (ty, nullable as)
+         pairToType mps backend (ty, nullable as)
         )
-    nameG = mkName $ unpack $ unHaskellName (entityHaskell t) ++ suffix
+    (nameFinal, paramsFinal)
+        | mpsGeneric mps = (nameG, [PlainTV backend])
+        | otherwise = (name, [])
+    nameG = mkName $ unpack $ unHaskellName (entityHaskell t) ++ "Generic"
     name = mkName $ unpack $ unHaskellName $ entityHaskell t
     cols = map (mkCol $ unHaskellName $ entityHaskell t) $ entityFields t
     backend = mkName "backend"
 
+    constrs
+        | entitySum t = map sumCon $ entityFields t
+        | otherwise = [RecC name cols]
+
+    sumCon fd@(FieldDef _ _ ty _) = NormalC
+        (sumConstrName t fd)
+        [(NotStrict, pairToType mps backend (ty, NotNullable))]
+
+sumConstrName :: EntityDef -> FieldDef -> Name
+sumConstrName t (FieldDef n _ _ _) = mkName $ unpack $ concat
+    [ unHaskellName $ entityHaskell t
+    , upperFirst $ unHaskellName n
+    , "Sum"
+    ]
+
 readMay :: Read a => String -> Maybe a
 readMay s =
     case reads s of
         (x, _):_ -> Just x
         [] -> Nothing
 
-entityUpdates :: EntityDef -> [(HaskellName, FieldType, Bool, PersistUpdate)]
+entityUpdates :: EntityDef -> [(HaskellName, FieldType, IsNullable, PersistUpdate)]
 entityUpdates =
     concatMap go . entityFields
   where
     go (FieldDef x _ y as) = map (\a -> (x, y, nullable as, a)) [minBound..maxBound]
 
-uniqueTypeDec :: EntityDef -> Dec
-uniqueTypeDec t =
+uniqueTypeDec :: MkPersistSettings -> EntityDef -> Dec
+uniqueTypeDec mps t =
     DataInstD [] ''Unique
-        [ ConT (mkName $ unpack (unHaskellName (entityHaskell t) ++ suffix))
-          `AppT` VarT backend, VarT backend2
-        ]
-            (map (mkUnique backend t) $ entityUniques t)
+        [genericDataType mps (unHaskellName $ entityHaskell t) $ VarT backend]
+            (map (mkUnique mps backend t) $ entityUniques t)
             []
   where
     backend = mkName "backend"
-    backend2 = mkName "backend2"
 
-mkUnique :: Name -> EntityDef -> UniqueDef -> Con
-mkUnique backend t (UniqueDef (HaskellName constr) _ fields) =
+mkUnique :: MkPersistSettings -> Name -> EntityDef -> UniqueDef -> Con
+mkUnique mps backend t (UniqueDef (HaskellName constr) _ fields attrs) =
     NormalC (mkName $ unpack constr) types
   where
     types = map (go . flip lookup3 (entityFields t))
           $ map (unHaskellName . fst) fields
 
-    go :: (FieldType, Bool) -> (Strict, Type)
-    go (_, True) = error "Error: cannot have nullables in unique"
-    go (ft, y) = (NotStrict, pairToType backend (ft, y))
+    force = "!force" `elem` attrs
 
-    lookup3 :: Text -> [FieldDef] -> (FieldType, Bool)
+    go :: (FieldType, IsNullable) -> (Strict, Type)
+    go (_, Nullable _) | not force = error nullErrMsg
+    go (ft, y) = (NotStrict, pairToType mps backend (ft, y))
+
+    lookup3 :: Text -> [FieldDef] -> (FieldType, IsNullable)
     lookup3 s [] =
         error $ unpack $ "Column not found: " ++ s ++ " in unique " ++ constr
     lookup3 x ((FieldDef (HaskellName x') _ y z):rest)
         | x == x' = (y, nullable z)
         | otherwise = lookup3 x rest
 
-pairToType :: Name -- ^ backend
-           -> (FieldType, Bool) -- ^ True == has Maybe attr
-           -> Type
-pairToType backend (s, False) = idType backend s
-pairToType backend (s, True) = ConT (mkName "Maybe") `AppT` idType backend s
+    nullErrMsg =
+      mconcat [ "Error:  By default we disallow NULLables in an uniqueness "
+              , "constraint.  The semantics of how NULL interacts with those "
+              , "constraints is non-trivial:  two NULL values are not "
+              , "considered equal for the purposes of an uniqueness "
+              , "constraint.  If you understand this feature, it is possible "
+              , "to use it your advantage.    *** Use a \"!force\" attribute "
+              , "on the end of the line that defines your uniqueness "
+              , "constraint in order to disable this check. ***" ]
 
-idType :: Name -> FieldType -> Type
-idType backend typ =
+pairToType :: MkPersistSettings
+           -> Name -- ^ backend
+           -> (FieldType, IsNullable)
+           -> Type
+pairToType mps backend (s, Nullable ByMaybeAttr) =
+  ConT (mkName "Maybe") `AppT` idType mps backend s
+pairToType mps backend (s, _) = idType mps backend s
+
+backendDataType :: MkPersistSettings -> Type
+backendDataType mps
+    | mpsGeneric mps = VarT $ mkName "backend"
+    | otherwise = mpsBackend mps
+
+genericDataType :: MkPersistSettings
+                -> Text -- ^ entity name
+                -> Type -- ^ backend
+                -> Type
+genericDataType mps typ' backend
+    | mpsGeneric mps = ConT (mkName $ unpack $ typ' ++ "Generic") `AppT` backend
+    | otherwise = ConT $ mkName $ unpack typ'
+
+idType :: MkPersistSettings -> Name -> FieldType -> Type
+idType mps backend typ =
     case stripId typ of
         Just typ' ->
-            ConT ''Key
-            `AppT` VarT backend
-            `AppT` (ConT (mkName $ unpack $ typ' ++ "Generic") `AppT` VarT backend)
+            ConT ''KeyBackend
+            `AppT` backend'
+            `AppT` genericDataType mps typ' (VarT backend)
         Nothing -> ftToType typ
+  where
+    backend'
+        | mpsGeneric mps = VarT backend
+        | otherwise = mpsBackend mps
 
 degen :: [Clause] -> [Clause]
 degen [] =
@@ -211,25 +276,48 @@ degen [] =
      in [Clause [WildP] (NormalB err) []]
 degen x = x
 
-mkToPersistFields :: [(String, Int)] -> Q Dec
-mkToPersistFields pairs = do
-    clauses <- mapM go pairs
-    return $ FunD (mkName "toPersistFields") $ degen clauses
+mkToPersistFields :: String -> EntityDef -> Q Dec
+mkToPersistFields constr ed@EntityDef { entitySum = isSum, entityFields = fields } = do
+    clauses <-
+        if isSum
+            then sequence $ zipWith goSum fields [1..]
+            else fmap return go
+    return $ FunD (mkName "toPersistFields") clauses
   where
-    go :: (String, Int) -> Q Clause
-    go (constr, fields) = do
-        xs <- sequence $ replicate fields $ newName "x"
+    go :: Q Clause
+    go = do
+        xs <- sequence $ replicate fieldCount $ newName "x"
         let pat = ConP (mkName constr) $ map VarP xs
         sp <- [|SomePersistField|]
         let bod = ListE $ map (AppE sp . VarE) xs
         return $ Clause [pat] (NormalB bod) []
 
+    fieldCount = length fields
+
+    goSum :: FieldDef -> Int -> Q Clause
+    goSum fd idx = do
+        let name = sumConstrName ed fd
+        enull <- [|SomePersistField PersistNull|]
+        let beforeCount = idx - 1
+            afterCount = fieldCount - idx
+            before = replicate beforeCount enull
+            after = replicate afterCount enull
+        x <- newName "x"
+        sp <- [|SomePersistField|]
+        let body = NormalB $ ListE $ mconcat
+                [ before
+                , [sp `AppE` VarE x]
+                , after
+                ]
+        return $ Clause [ConP name [VarP x]] body []
+
+
 mkToFieldNames :: [UniqueDef] -> Q Dec
 mkToFieldNames pairs = do
     pairs' <- mapM go pairs
     return $ FunD (mkName "persistUniqueToFieldNames") $ degen pairs'
   where
-    go (UniqueDef constr _ names) = do
+    go (UniqueDef constr _ names _) = do
         names' <- lift names
         return $
             Clause
@@ -252,7 +340,7 @@ mkUniqueToValues pairs = do
     return $ FunD (mkName "persistUniqueToValues") $ degen pairs'
   where
     go :: UniqueDef -> Q Clause
-    go (UniqueDef constr _ names) = do
+    go (UniqueDef constr _ names _) = do
         xs <- mapM (const $ newName "x") names
         let pat = ConP (mkName $ unpack $ unHaskellName constr) $ map VarP xs
         tpv <- [|toPersistValue|]
@@ -282,22 +370,26 @@ mkToValue func = FunD (mkName func) . degen . map go
                    (NormalB $ VarE (mkName "toPersistValue") `AppE` VarE x)
                    []
 
-mkHalfDefined :: String -> Int -> Dec
+mkHalfDefined :: Name -> Int -> Dec
 mkHalfDefined constr count' =
         FunD (mkName "halfDefined")
             [Clause [] (NormalB
-            $ foldl AppE (ConE $ mkName constr)
+            $ foldl AppE (ConE constr)
                     (replicate count' $ VarE $ mkName "undefined")) []]
 
+isNotNull :: PersistValue -> Bool
+isNotNull PersistNull = False
+isNotNull _ = True
+
 mkFromPersistValues :: EntityDef -> Q [Clause]
-mkFromPersistValues t = do
+mkFromPersistValues t@(EntityDef { entitySum = False }) = do
     nothing <- [|Left $(liftT "Invalid fromPersistValues input")|]
     let cons' = ConE $ mkName $ unpack $ unHaskellName $ entityHaskell t
     xs <- mapM (const $ newName "x") $ entityFields t
     fs <- [|fromPersistValue|]
     let xs' = map (AppE fs . VarE) xs
     let pat = ListP $ map VarP xs
-    ap' <- [|apE|]
+    ap' <- [|(<*>)|]
     just <- [|Right|]
     let cons'' = just `AppE` cons'
     return
@@ -306,20 +398,99 @@ mkFromPersistValues t = do
         ]
   where
     go ap' x y = InfixE (Just x) ap' (Just y)
+mkFromPersistValues t@(EntityDef { entitySum = True }) = do
+    nothing <- [|Left $(liftT "Invalid fromPersistValues input: sum type with all nulls")|]
+    clauses <- mkClauses [] $ entityFields t
+    return $ clauses `mappend` [Clause [WildP] (NormalB nothing) []]
+  where
+    mkClauses _ [] = return []
+    mkClauses before (field:after) = do
+        x <- newName "x"
+        let null' = ConP 'PersistNull []
+            pat = ListP $ mconcat
+                [ map (const null') before
+                , [VarP x]
+                , map (const null') after
+                ]
+            constr = ConE $ sumConstrName t field
+        fmap' <- [|fmap|]
+        fs <- [|fromPersistValue $(return $ VarE x)|]
+        let guard' = NormalG $ VarE 'isNotNull `AppE` VarE x
+        let clause = Clause [pat] (GuardedB [(guard', InfixE (Just constr) fmap' (Just fs))]) []
+        clauses <- mkClauses (field : before) after
+        return $ clause : clauses
 
+type Lens s t a b = forall f. Functor f => (a -> f b) -> s -> f t
+
+lens :: (s -> a) -> (s -> b -> t) -> Lens s t a b
+lens sa sbt afb s = fmap (sbt s) (afb $ sa s)
+
+mkLensClauses :: EntityDef -> Q [Clause]
+mkLensClauses t = do
+    lens' <- [|lens|]
+    getId <- [|entityKey|]
+    setId <- [|\(Entity _ val) key -> Entity key val|]
+    getVal <- [|entityVal|]
+    dot <- [|(.)|]
+    keyName <- newName "key"
+    valName <- newName "val"
+    xName <- newName "x"
+    let idClause = Clause
+            [ConP (mkName $ unpack $ unHaskellName (entityHaskell t) ++ "Id") []]
+            (NormalB $ lens' `AppE` getId `AppE` setId)
+            []
+    if entitySum t
+        then return $ idClause : map (toSumClause lens' getVal dot keyName valName xName) (entityFields t)
+        else return $ idClause : map (toClause lens' getVal dot keyName valName xName) (entityFields t)
+  where
+    toClause lens' getVal dot keyName valName xName f = Clause
+        [ConP (mkName $ unpack $ unHaskellName (entityHaskell t) ++ upperFirst (unHaskellName $ fieldHaskell f)) []]
+        (NormalB $ lens' `AppE` getter `AppE` setter)
+        []
+      where
+        fieldName = mkName $ unpack $ recName (unHaskellName $ entityHaskell t) (unHaskellName $ fieldHaskell f)
+        getter = InfixE (Just $ VarE fieldName) dot (Just getVal)
+        setter = LamE
+            [ ConP 'Entity [VarP keyName, VarP valName]
+            , VarP xName
+            ]
+            $ ConE 'Entity `AppE` VarE keyName `AppE` RecUpdE
+                (VarE valName)
+                [(fieldName, VarE xName)]
+
+    toSumClause lens' getVal dot keyName valName xName f = Clause
+        [ConP (mkName $ unpack $ unHaskellName (entityHaskell t) ++ upperFirst (unHaskellName $ fieldHaskell f)) []]
+        (NormalB $ lens' `AppE` getter `AppE` setter)
+        []
+      where
+        fieldName = mkName $ unpack $ recName (unHaskellName $ entityHaskell t) (unHaskellName $ fieldHaskell f)
+        getter = LamE
+            [ ConP 'Entity [WildP, VarP valName]
+            ] $ CaseE (VarE valName)
+            [ Match (ConP (sumConstrName t f) [VarP xName]) (NormalB $ VarE xName) []
+
+            -- FIXME It would be nice if the types expressed that the Field is
+            -- a sum type and therefore could result in Maybe.
+            , Match WildP (NormalB $ VarE 'error `AppE` LitE (StringL "Tried to use fieldLens on a Sum type")) []
+            ]
+        setter = LamE
+            [ ConP 'Entity [VarP keyName, WildP]
+            , VarP xName
+            ]
+            $ ConE 'Entity `AppE` VarE keyName `AppE` (ConE (sumConstrName t f) `AppE` VarE xName)
 
 mkEntity :: MkPersistSettings -> EntityDef -> Q [Dec]
 mkEntity mps t = do
     t' <- lift t
     let nameT = unHaskellName $ entityHaskell t
     let nameS = unpack nameT
-    let clazz = ConT ''PersistEntity `AppT` (ConT (mkName $ unpack $ unHaskellName (entityHaskell t) ++ suffix) `AppT` VarT (mkName "backend"))
-    tpf <- mkToPersistFields [(nameS, length $ entityFields t)]
+    let clazz = ConT ''PersistEntity `AppT` genericDataType mps (unHaskellName $ entityHaskell t) (VarT $ mkName "backend")
+    tpf <- mkToPersistFields nameS t
     fpv <- mkFromPersistValues t
     utv <- mkUniqueToValues $ entityUniques t
     puk <- mkUniqueKeys t
 
-    fields <- mapM (mkField t) $ FieldDef
+    fields <- mapM (mkField mps t) $ FieldDef
         (HaskellName "Id")
         (entityID t)
         (FTTypeCon Nothing $ unHaskellName (entityHaskell t) ++ "Id")
@@ -327,26 +498,35 @@ mkEntity mps t = do
         : entityFields t
     toFieldNames <- mkToFieldNames $ entityUniques t
 
-    return
-      [ dataTypeDec t
-      , TySynD (mkName nameS) [] $
-            ConT (mkName $ unpack $ nameT ++ suffix)
-                `AppT` mpsBackend mps
+    let addSyn -- FIXME maybe remove this
+            | mpsGeneric mps = (:) $
+                TySynD (mkName nameS) [] $
+                    genericDataType mps nameT $ mpsBackend mps
+            | otherwise = id
+
+    lensClauses <- mkLensClauses t
+
+    return $ addSyn
+      [ dataTypeDec mps t
       , TySynD (mkName $ unpack $ unHaskellName (entityHaskell t) ++ "Id") [] $
-            ConT ''Key `AppT` mpsBackend mps `AppT` ConT (mkName nameS)
+            ConT ''KeyBackend `AppT` mpsBackend mps `AppT` ConT (mkName nameS)
       , InstanceD [] clazz $
-        [ uniqueTypeDec t
+        [ uniqueTypeDec mps t
         , FunD (mkName "entityDef") [Clause [WildP] (NormalB t') []]
         , tpf
         , FunD (mkName "fromPersistValues") fpv
-        , mkHalfDefined nameS $ length $ entityFields t
+        , mkHalfDefined
+            (if entitySum t
+                then sumConstrName t (head $ entityFields t)
+                else mkName nameS)
+            (if entitySum t then 1 else length $ entityFields t)
         , toFieldNames
         , utv
         , puk
         , DataInstD
             []
             ''EntityField
-            [ ConT (mkName $ unpack $ nameT ++ suffix) `AppT` VarT (mkName "backend")
+            [ genericDataType mps nameT $ VarT $ mkName "backend"
             , VarT $ mkName "typ"
             ]
             (map fst fields)
@@ -354,9 +534,10 @@ mkEntity mps t = do
         , FunD (mkName "persistFieldDef") (map snd fields)
         , TySynInstD
             (mkName "PersistEntityBackend")
-            [ConT (mkName $ unpack $ unHaskellName (entityHaskell t) ++ suffix) `AppT` VarT (mkName "backend")]
-            (VarT (mkName "backend"))
+            [genericDataType mps (unHaskellName $ entityHaskell t) $ VarT $ mkName "backend"]
+            (backendDataType mps)
         , FunD (mkName "persistIdField") [Clause [] (NormalB $ ConE $ mkName $ unpack $ unHaskellName (entityHaskell t) ++ "Id") []]
+        , FunD (mkName "fieldLens") lensClauses
         ]
       ]
 
@@ -370,16 +551,15 @@ mkEntity mps t = do
 --            Right r -> r) o
 --    fromPersistValue x = Left $ "Expected PersistMap, received: " ++ show x
 --    sqlType _ = SqlString
-persistFieldFromEntity :: EntityDef -> Q [Dec]
-persistFieldFromEntity e = do
+persistFieldFromEntity :: MkPersistSettings -> EntityDef -> Q [Dec]
+persistFieldFromEntity mps e = do
     ss <- [|SqlString|]
     let columnNames = map (unpack . unHaskellName . fieldHaskell) (entityFields e)
     obj <- [|\ent -> PersistMap $ zip (map pack columnNames) (map toPersistValue $ toPersistFields ent)|]
     fpv <- [|\x -> fromPersistValues $ map (\(_,v) -> case fromPersistValue v of
                                                       Left e' -> error $ unpack e'
                                                       Right r -> r) x|]
-    let typ = ConT (mkName $ entityName `mappend` "Generic")
-              `AppT` VarT (mkName "backend")
+    let typ = genericDataType mps (pack entityName) $ VarT $ mkName "backend"
 
     compose <- [|(<=<)|]
     getPersistMap' <- [|getPersistMap|]
@@ -436,12 +616,12 @@ data Dep = Dep
     { depTarget :: Text
     , depSourceTable :: HaskellName
     , depSourceField :: HaskellName
-    , depSourceNull :: Bool
+    , depSourceNull  :: IsNullable
     }
 
 -- | Generate a 'DeleteCascade' instance for the given @EntityDef@s.
-mkDeleteCascade :: [EntityDef] -> Q [Dec]
-mkDeleteCascade defs = do
+mkDeleteCascade :: MkPersistSettings -> [EntityDef] -> Q [Dec]
+mkDeleteCascade mps defs = do
     let deps = concatMap getDeps defs
     mapM (go deps) defs
   where
@@ -451,22 +631,21 @@ mkDeleteCascade defs = do
       where
         getDeps' :: FieldDef -> [Dep]
         getDeps' (FieldDef name _ ftyp attribs) =
-            let isNull = nullable attribs
-             in case stripId ftyp of
-                    Just f ->
-                         return Dep
-                            { depTarget = f
-                            , depSourceTable = entityHaskell def
-                            , depSourceField = name
-                            , depSourceNull = isNull
-                            }
-                    Nothing -> []
+            case stripId ftyp of
+                Just f ->
+                     return Dep
+                        { depTarget = f
+                        , depSourceTable = entityHaskell def
+                        , depSourceField = name
+                        , depSourceNull  = nullable attribs
+                        }
+                Nothing -> []
     go :: [Dep] -> EntityDef -> Q Dec
     go allDeps EntityDef{entityHaskell = name} = do
         let deps = filter (\x -> depTarget x == unHaskellName name) allDeps
         key <- newName "key"
-        del <- [|delete|]
-        dcw <- [|deleteCascadeWhere|]
+        let del = VarE 'delete
+        let dcw = VarE 'deleteCascadeWhere
         just <- [|Just|]
         filt <- [|Filter|]
         eq <- [|Eq|]
@@ -482,29 +661,30 @@ mkDeleteCascade defs = do
               where
                 filtName = unHaskellName (depSourceTable dep) ++
                            upperFirst (unHaskellName $ depSourceField dep)
-                val False = VarE key
-                val True = just `AppE` VarE key
+                val (Nullable ByMaybeAttr) = just `AppE` VarE key
+                val _                      =             VarE key
 
 
 
         let stmts :: [Stmt]
             stmts = map mkStmt deps `mappend`
                     [NoBindS $ del `AppE` VarE key]
+
+        let entityT = genericDataType mps (unHaskellName name) $ VarT $ mkName "backend"
+
         return $
             InstanceD
-            [ ClassP ''PersistQuery [VarT $ mkName "backend", VarT $ mkName "m"]
-            , ClassP ''Monad [VarT $ mkName "m"]
+            [ ClassP ''PersistQuery [VarT $ mkName "m"]
+            , EqualP (ConT ''PersistEntityBackend `AppT` entityT) (ConT ''PersistMonadBackend `AppT` VarT (mkName "m"))
             ]
-            (ConT ''DeleteCascade `AppT`
-                (ConT (mkName $ unpack $ unHaskellName name ++ suffix) `AppT` VarT (mkName "backend"))
-                `AppT` VarT (mkName "backend")
-                `AppT` VarT (mkName "m")
-                )
+            (ConT ''DeleteCascade `AppT` entityT `AppT` VarT (mkName "m"))
             [ FunD (mkName "deleteCascade")
                 [Clause [VarP key] (NormalB $ DoE stmts) []]
             ]
 
 mkUniqueKeys :: EntityDef -> Q Dec
+mkUniqueKeys def | entitySum def =
+    return $ FunD (mkName "persistUniqueKeys") [Clause [WildP] (NormalB $ ListE []) []]
 mkUniqueKeys def = do
     c <- clause
     return $ FunD (mkName "persistUniqueKeys") [c]
@@ -520,7 +700,7 @@ mkUniqueKeys def = do
         return $ Clause [pat] (NormalB $ ListE pcs) []
 
     go :: [(HaskellName, Name)] -> UniqueDef -> Exp
-    go xs (UniqueDef name _ cols) =
+    go xs (UniqueDef name _ cols _) =
         foldl' (go' xs) (ConE (mkName $ unpack $ unHaskellName name)) (map fst cols)
 
     go' :: [(HaskellName, Name)] -> Exp -> HaskellName -> Exp
@@ -603,7 +783,7 @@ mkMigrate fun allDefs = do
     undefinedEntityTH u = SigE u . ConT . mkName . unpack . unHaskellName . entityHaskell
 
 instance Lift EntityDef where
-    lift (EntityDef a b c d e f g h) =
+    lift (EntityDef a b c d e f g h i) =
         [|EntityDef
             $(lift a)
             $(lift b)
@@ -613,11 +793,12 @@ instance Lift EntityDef where
             $(lift f)
             $(liftTs g)
             $(liftMap h)
+            $(lift i)
             |]
 instance Lift FieldDef where
     lift (FieldDef a b c d) = [|FieldDef $(lift a) $(lift b) $(lift c) $(liftTs d)|]
 instance Lift UniqueDef where
-    lift (UniqueDef a b c) = [|UniqueDef $(lift a) $(lift b) $(lift c)|]
+    lift (UniqueDef a b c d) = [|UniqueDef $(lift a) $(lift b) $(lift c) $(liftTs d)|]
 
 pack' :: String -> Text
 pack' = pack
@@ -674,8 +855,8 @@ instance Lift PersistUpdate where
 -- forall . typ ~ FieldType => EntFieldName
 --
 -- EntFieldName = FieldDef ....
-mkField :: EntityDef -> FieldDef -> Q (Con, Clause)
-mkField et cd = do
+mkField :: MkPersistSettings -> EntityDef -> FieldDef -> Q (Con, Clause)
+mkField mps et cd = do
     let con = ForallC
                 []
                 [EqualP (VarT $ mkName "typ") maybeTyp]
@@ -692,17 +873,17 @@ mkField et cd = do
         , upperFirst $ unHaskellName $ fieldHaskell cd
         ]
     maybeTyp =
-        if nullable $ fieldAttrs cd
+        if nullable (fieldAttrs cd) == Nullable ByMaybeAttr
             then ConT ''Maybe `AppT` typ
             else typ
     typ =
         case stripId $ fieldType cd of
             Just ft ->
-                 ConT ''Key
-                    `AppT` (VarT $ mkName "backend")
-                    `AppT`
-                        let con = ConT $ mkName $ unpack $ ft ++ suffix
-                         in con `AppT` VarT (mkName "backend")
+                 ConT ''KeyBackend
+                    `AppT` (if mpsGeneric mps
+                                then VarT $ mkName "backend"
+                                else mpsBackend mps)
+                    `AppT` genericDataType mps ft (VarT $ mkName "backend")
             Nothing -> ftToType $ fieldType cd
 
 ftToType :: FieldType -> Type
@@ -711,16 +892,13 @@ ftToType (FTTypeCon (Just m) t) = ConT $ mkName $ unpack $ concat [m, ".", t]
 ftToType (FTApp x y) = ftToType x `AppT` ftToType y
 ftToType (FTList x) = ListT `AppT` ftToType x
 
-suffix :: Text
-suffix = "Generic"
-
 infixr 5 ++
 (++) :: Text -> Text -> Text
 (++) = append
 
-mkJSON :: EntityDef -> Q [Dec]
-mkJSON def | not ("json" `elem` entityAttrs def) = return []
-mkJSON def = do
+mkJSON :: MkPersistSettings -> EntityDef -> Q [Dec]
+mkJSON _ def | not ("json" `elem` entityAttrs def) = return []
+mkJSON mps def = do
     pureE <- [|pure|]
     apE' <- [|(<*>)|]
     packE <- [|pack|]
@@ -734,10 +912,8 @@ mkJSON def = do
     xs <- mapM (newName . unpack . unHaskellName . fieldHaskell)
         $ entityFields def
 
-    let con = ConT $ mkName $ unpack
-              (unHaskellName (entityHaskell def) ++ "Generic")
-        conName = mkName $ unpack $ unHaskellName $ entityHaskell def
-        typ = con `AppT` VarT (mkName "backend")
+    let conName = mkName $ unpack $ unHaskellName $ entityHaskell def
+        typ = genericDataType mps (unHaskellName $ entityHaskell def) $ VarT $ mkName "backend"
         toJSONI = InstanceD
             []
             (ConT ''ToJSON `AppT` typ)
@@ -768,6 +944,6 @@ mkJSON def = do
         pulls = map toPull $ entityFields def
         toPull f = InfixE
             (Just $ VarE obj)
-            (if nullable (fieldAttrs f) then dotColonQE else dotColonE)
+            (if nullable (fieldAttrs f) == Nullable ByMaybeAttr then dotColonQE else dotColonE)
             (Just $ AppE packE $ LitE $ StringL $ unpack $ unHaskellName $ fieldHaskell f)
     return [toJSONI, fromJSONI]
