@@ -1,19 +1,19 @@
 {-# LANGUAGE OverloadedStrings, QuasiQuotes, TemplateHaskell, TypeFamilies, GADTs #-}
 {-# LANGUAGE EmptyDataDecls #-}
-import Test.Hspec.Monadic
+{-# LANGUAGE FlexibleInstances, MultiParamTypeClasses #-}
+import Test.Hspec
 import Test.Hspec.QuickCheck
-import Test.Hspec.HUnit()
-import Test.HUnit
 import Data.ByteString.Lazy.Char8 ()
 import Test.QuickCheck.Arbitrary
 import Control.Applicative ((<$>), (<*>))
 
 import Database.Persist
+import Database.Persist.Store (DeleteCascade (..))
 import Database.Persist.TH
 import Data.Text (Text, pack)
 import Data.Aeson
 
-mkPersist sqlSettings [persistUpperCase|
+share [mkPersist sqlSettings { mpsGeneric = False }, mkDeleteCascade sqlSettings { mpsGeneric = False }] [persistUpperCase|
 Person json
     name Text
     age Int Maybe
@@ -30,23 +30,23 @@ NoJson
 |]
 
 -- ensure no-json works
-instance ToJSON (NoJsonGeneric b) where
+instance ToJSON NoJson where
     toJSON = undefined
-instance FromJSON (NoJsonGeneric b) where
+instance FromJSON NoJson where
     parseJSON = undefined
 
 arbitraryT = pack <$> arbitrary
 
-instance Arbitrary (PersonGeneric b) where
+instance Arbitrary Person where
     arbitrary = Person <$> arbitraryT <*> arbitrary <*> arbitrary
-instance Arbitrary (AddressGeneric b) where
+instance Arbitrary Address where
     arbitrary = Address <$> arbitraryT <*> arbitraryT <*> arbitrary
 
 main :: IO ()
-main = hspecX $ do
+main = hspec $ do
     describe "JSON serialization" $ do
         prop "to/from is idempotent" $ \person ->
             decode (encode person) == Just (person :: Person)
         it "decode" $
-            decode "{\"name\":\"Michael\",\"age\":27,\"address\":{\"street\":\"Narkis\",\"city\":\"Maalot\"}}" @?= Just
+            decode "{\"name\":\"Michael\",\"age\":27,\"address\":{\"street\":\"Narkis\",\"city\":\"Maalot\"}}" `shouldBe` Just
                 (Person "Michael" (Just 27) $ Address "Narkis" "Maalot" Nothing)
